from configurations import values

from ._base import ConfigMixin


class CeleryMixin(ConfigMixin):
    """
    Configure Celery.

    The Celery app must be constructed as:
        app = Celery(
            '<my_project_name>',  # Change this
            config_source='django.conf:settings',
            namespace='CELERY'
        )

    The `DJANGO_CELERY_BROKER_URL` environment variable may be externally set to an AMQP URL.
    """

    # Assume AMQP.
    CELERY_BROKER_URL = values.Value('amqp://localhost:5672/')

    # Disable results backend, as this feature has too many weaknesses.
    # The database should be used to communicate results of completed tasks.
    CELERY_RESULT_BACKEND = None

    # Only acknowledge a task being done after the function finishes.
    # This provides safety against worker crashes, but adds the requirement
    # that tasks must be idempotent (which is a best practice anyway).
    # See: https://docs.celeryproject.org/en/stable/faq.html#should-i-use-retry-or-acks-late
    # Acknowledge early in development, which will help prevent failing or
    # long-running tasks from being started automatically every time the worker
    # process restarts; this more aggressively flushes the task queue.
    @property
    def CELERY_TASK_ACKS_LATE(self):  # noqa: N802
        return False if self.DEBUG else True

    # CloudAMQP-suggested settings
    # https://www.cloudamqp.com/docs/celery.html
    CELERY_BROKER_POOL_LIMIT = 1
    CELERY_BROKER_HEARTBEAT = None
    CELERY_BROKER_CONNECTION_TIMEOUT = 30
    CELERY_EVENT_QUEUE_EXPIRES = 60

    # Note, CELERY_WORKER settings could be different on each running worker.

    # Do not prefetch, as the speed benefit for fast-running tasks may not be
    # worth a potentially unfair allocation with slow-running tasks and
    # multiple workers.
    CELERY_WORKER_PREFETCH_MULTIPLIER = 1

    # In development, run without concurrency, otherwise accept the default of
    # the number of CPU cores.
    # Workers running memory-intensive tasks may need to decrease this.
    @property
    def CELERY_WORKER_CONCURRENCY(self):  # noqa: N802
        return 1 if self.DEBUG else None



from importlib.metadata import PackageNotFoundError, version

from ._allauth import AllauthMixin
from ._base import ComposedConfiguration, ConfigMixin
from ._celery import CeleryMixin
from ._configuration import (
    DevelopmentBaseConfiguration,
    HerokuProductionBaseConfiguration,
    ProductionBaseConfiguration,
    TestingBaseConfiguration,
)
from ._cors import CorsMixin
from ._database import DatabaseMixin
from ._debug import DebugMixin
from ._django import DjangoMixin
from ._email import ConsoleEmailMixin, SmtpEmailMixin
from ._extensions import ExtensionsMixin
from ._filter import FilterMixin
from ._https import HttpsMixin
from ._logging import LoggingMixin
from ._rest_framework import RestFrameworkMixin
from ._sentry import SentryMixin
from ._static import StaticFileMixin, WhitenoiseStaticFileMixin
from ._storage import MinioStorageMixin, S3StorageMixin

__all__ = [
    'AllauthMixin',
    'CeleryMixin',
    'ComposedConfiguration',
    'ConfigMixin',
    'ConsoleEmailMixin',
    'CorsMixin',
    'DatabaseMixin',
    'DebugMixin',
    'DevelopmentBaseConfiguration',
    'DjangoMixin',
    'ExtensionsMixin',
    'FilterMixin',
    'HerokuProductionBaseConfiguration',
    'HttpsMixin',
    'LoggingMixin',
    'MinioStorageMixin',
    'ProductionBaseConfiguration',
    'RestFrameworkMixin',
    'S3StorageMixin',
    'SentryMixin',
    'SmtpEmailMixin',
    'StaticFileMixin',
    'TestingBaseConfiguration',
    'WhitenoiseStaticFileMixin',
]

try:
    __version__ = version('django-composed-configuration')
except PackageNotFoundError:
    # package is not installed
    pass



from typing import Any, Dict, Type

from ._base import ComposedConfiguration, ConfigMixin


class RestFrameworkMixin(ConfigMixin):
    """
    Configure Django REST Framework.

    This requires the `django-cors-headers`, `django-girder-utils`, `django-oauth-toolkit`,
    and `drf-yasg` packages to be installed.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        configuration.INSTALLED_APPS += [
            'rest_framework',
            'rest_framework.authtoken',
            'oauth2_provider',
            'drf_yasg',
        ]

        if configuration.DEBUG:
            configuration.OAUTH2_PROVIDER['ALLOWED_REDIRECT_URI_SCHEMES'] = ['http', 'https']
            # In development, always present the approval dialog
            configuration.OAUTH2_PROVIDER['REQUEST_APPROVAL_PROMPT'] = 'force'

    # When SessionAuthentication is allowed, it's critical that the following settings
    # (respectively part of Django and django-cors-headers) are set to these values (although those
    # are the also the default values).
    SESSION_COOKIE_SAMESITE = 'Lax'
    CORS_ALLOW_CREDENTIALS = False

    REST_FRAMEWORK = {
        'DEFAULT_AUTHENTICATION_CLASSES': [
            'oauth2_provider.contrib.rest_framework.OAuth2Authentication',
            # Allow SessionAuthentication, as this is much more convenient for Ajax requests
            # from server-rendered pages, including:
            # * YASG (Swagger / ReDoc)
            # * The Admin interface, when using interactive fields like S3-file-field
            # * Augmentation of server-rendered views with background Javascript
            #   (see https://docs.djangoproject.com/en/3.1/ref/csrf/#ajax )
            # It's important that true SPAs and other clients be forced to go though
            # OAuth2Authentication instead, as this is the only supported auth mechanism which
            # robustly works across origins; however, it turns out that this can only be enforced
            # partially.
            # To understand why, first read https://web.dev/same-site-same-origin/ to understand
            # that even with "SameSize=Lax" (or "SameSite=Strict"), cookies are only technically
            # limited to same-site requests, and do not have the stronger same-origin limitation.
            # If a naive SPA developer configures their client to include credentials
            # ("{withCredentials: true}" in XHR, jQuery, and Axios, or "{credentials: 'include'}"
            # in Fetch; this configuration is often suggested as an "easy" fix for authentication
            # problems by StackOverflow), then the session cookie will be sent with any cross-site
            # requests where the user has logged into the Django server. Note, cross-site requests
            # include origins with a different port (typical in local development) and origins with
            # a different subdomain (common in many deployments). From DRF's perspective, as long
            # as the request uses a safe verb (more on this below), the request will be
            # authenticated transparently. However, since Django is configured to not set
            # "Access-Control-Allow-Credentials", the SPA client will not be able to read the
            # response and get a CORS error; this is the right outcome (client cannot effectively
            # make the request), but with a confusing and hard-to-debug reason (CORS error,
            # instead of a 401/403) and developers may be tempted to "fix" it by enabling
            # "Access-Control-Allow-Credentials" instead of fixing their client to use OAuth
            # correctly, which will likely lead to further bugs when the SPA is deployed to a
            # non-same-site environment. Alternatively, if the request does not use a safe verb
            # (and the request is not preflighted by the browser, which is permitted in some
            # cases), DRF will enforce CSRF protection and the request will 403 fail with a
            # "CSRF Failed: CSRF token missing or incorrect" message, which is also confusing and
            # may lead developers to incorrect fixes.
            # TL;DR: Developers of SPAs may encounter misleading error messages when making Ajax
            # requests "withCredentials", but security is still maintained.
            'rest_framework.authentication.SessionAuthentication',
        ],
        # This is a much more sensible degree of basic security
        'DEFAULT_PERMISSION_CLASSES': ['rest_framework.permissions.IsAuthenticatedOrReadOnly'],
        # BoundedLimitOffsetPagination provides LimitOffsetPagination with a maximum page size
        'DEFAULT_PAGINATION_CLASS': 'girder_utils.rest_framework.BoundedLimitOffsetPagination',
        # This provides a sane default for requests that do not specify a page size.
        # This also ensures that endpoints with pagination will always return a
        # pagination-structured response.
        'PAGE_SIZE': 100,
        # Real clients typically JSON-encode their request bodies, so the test client should too
        'TEST_REQUEST_DEFAULT_FORMAT': 'json',
    }

    OAUTH2_PROVIDER = {
        'PKCE_REQUIRED': True,
        'ALLOWED_REDIRECT_URI_SCHEMES': ['https'],
        # Don't require users to re-approve scopes each time
        'REQUEST_APPROVAL_PROMPT': 'auto',
        # ERROR_RESPONSE_WITH_SCOPES is only used with the "permission_classes" helpers for scopes.
        # If the scope itself is confidential, this could leak information. but the usability
        # benefit is probably worth it.
        'ERROR_RESPONSE_WITH_SCOPES': True,
        # Django can persist logins for longer than this via cookies,
        # but non-refreshing clients will need to redirect to Django's auth every 24 hours.
        'ACCESS_TOKEN_EXPIRE_SECONDS': 24 * 60 * 60,
        # This allows refresh tokens to eventually be removed from the database by
        # "manage.py cleartokens". This value is not actually enforced when refresh tokens are
        # checked, but it can be assumed that all clients will need to redirect to Django's auth
        # every 30 days.
        'REFRESH_TOKEN_EXPIRE_SECONDS': 30 * 24 * 60 * 60,
    }

    SWAGGER_SETTINGS: Dict[str, Any] = {
        # The default security definition ("basic") is not supported by this DRF configuration,
        # so expect all logins to come via the Django session, which there's no OpenAPI
        # security definition for.
        'SECURITY_DEFINITIONS': None,
        'USE_SESSION_AUTH': True,
    }

    REDOC_SETTINGS: Dict[str, Any] = {}



from typing import List, Optional, Tuple

from configurations import values

from ._allauth import AllauthMixin
from ._base import ComposedConfiguration
from ._celery import CeleryMixin
from ._cors import CorsMixin
from ._database import DatabaseMixin
from ._debug import DebugMixin
from ._django import DjangoMixin
from ._docker import _AlwaysContains, _is_docker
from ._email import ConsoleEmailMixin, SmtpEmailMixin
from ._extensions import ExtensionsMixin
from ._filter import FilterMixin
from ._https import HttpsMixin
from ._logging import LoggingMixin
from ._rest_framework import RestFrameworkMixin
from ._sentry import SentryMixin
from ._static import WhitenoiseStaticFileMixin
from ._storage import MinioStorageMixin, S3StorageMixin


# Subclasses are loaded in last to first ordering
class _BaseConfiguration(
    ExtensionsMixin,
    CeleryMixin,
    RestFrameworkMixin,
    FilterMixin,
    # CorsMixin must be loaded after WhitenoiseStaticFileMixin
    CorsMixin,
    WhitenoiseStaticFileMixin,
    DatabaseMixin,
    LoggingMixin,
    AllauthMixin,
    # DjangoMixin should be loaded first
    DjangoMixin,
    ComposedConfiguration,
):
    pass


class DevelopmentBaseConfiguration(
    DebugMixin, ConsoleEmailMixin, MinioStorageMixin, _BaseConfiguration
):
    DEBUG = True
    SECRET_KEY = 'insecuresecret'

    ALLOWED_HOSTS = values.ListValue(['localhost', '127.0.0.1'])
    CORS_ORIGIN_REGEX_WHITELIST = values.ListValue(
        [r'^https?://localhost:\d+$', r'^https?://127\.0\.0\.1:\d+$']
    )

    # When in Docker, the bridge network sends requests from the host machine exclusively via a
    # dedicated IP address. Since there's no way to determine the real origin address,
    # consider any IP address (though actually this will only be the single dedicated address) to
    # be internal. This relies on the host to set up appropriate firewalls for Docker, to prevent
    # access from non-internal addresses.
    INTERNAL_IPS = _AlwaysContains() if _is_docker() else ['127.0.0.1']
    # Setting this allows MinIO to work through network namespace partitions
    # (e.g. when running within Docker Compose)
    MINIO_STORAGE_MEDIA_URL = values.Value(None)


class TestingBaseConfiguration(MinioStorageMixin, _BaseConfiguration):
    SECRET_KEY = 'testingsecret'

    # Testing will add 'testserver' to ALLOWED_HOSTS
    ALLOWED_HOSTS: List[str] = []

    MINIO_STORAGE_MEDIA_BUCKET_NAME = 'test-django-storage'

    # Testing will set EMAIL_BACKEND to use the memory backend


class ProductionBaseConfiguration(
    SentryMixin, SmtpEmailMixin, S3StorageMixin, HttpsMixin, _BaseConfiguration
):
    pass


# Separate these settings for potential reuse, but this may not be usable enough to make public
# yet, given the fragility of ensuring that this is added as a superclass in the correct order.
class _HerokuMixin:
    # Use different env var names (with no DJANGO_ prefix) for services that Heroku auto-injects
    DATABASES = values.DatabaseURLValue(
        environ_name='DATABASE_URL',
        environ_prefix=None,
        environ_required=True,
        engine='django.db.backends.postgresql',
        conn_max_age=600,
        ssl_require=True,
    )
    CELERY_BROKER_URL = values.Value(
        environ_name='CLOUDAMQP_URL', environ_prefix=None, environ_required=True
    )
    # https://help.heroku.com/J2R1S4T8/can-heroku-force-an-application-to-use-ssl-tls
    SECURE_PROXY_SSL_HEADER: Optional[Tuple[str, str]] = ('HTTP_X_FORWARDED_PROTO', 'https')


class HerokuProductionBaseConfiguration(_HerokuMixin, ProductionBaseConfiguration):
    pass



from typing import Optional, Tuple

from ._base import ConfigMixin


class HttpsMixin(ConfigMixin):
    """Configure Django's security middleware for HTTPS."""

    SECURE_SSL_REDIRECT = True

    # This must be set when deployed behind a proxy
    SECURE_PROXY_SSL_HEADER: Optional[Tuple[str, str]] = None

    SESSION_COOKIE_SECURE = True
    CSRF_COOKIE_SECURE = True

    # Enable HSTS
    SECURE_HSTS_SECONDS = 60 * 60 * 24 * 365  # 1 year
    # This is already False by default, but it's important to ensure HSTS is not forced on other
    # subdomains which may have different HTTPS practices.
    SECURE_HSTS_INCLUDE_SUBDOMAINS = False



from pathlib import Path
from typing import Type

from configurations import values

from ._base import ComposedConfiguration, ConfigMixin


class StaticFileMixin(ConfigMixin):
    """
    Configure static file collection by Django, with serving in development.

    This could be used directly, but is typically included implicitly by WhitenoiseStaticFileMixin.

    Downstreams must explicitly define the settings:
    * `BASE_DIR`, as the pathlib Path to the project root. This can typically be done as:
      `BASE_DIR = Path(__file__).resolve(strict=True).parent.parent`.
    """

    STATIC_URL = '/static/'

    # BASE_DIR is in Django's startproject template, but isn't actually used as a real setting
    @property
    def BASE_DIR(self) -> Path:  # noqa: N802
        raise Exception('BASE_DIR must be explicitly set to the path of the Django project root.')

    @property
    def STATIC_ROOT(self):  # noqa: N802
        # Django staticfiles creates any intermediate directories which don't exist
        # TODO: allow from env?
        return values.PathValue(
            str(Path(self.BASE_DIR) / 'staticfiles'),
            environ=False,
            check_exists=False,
            # Disable late_binding, to make this return an actual str, not a Value, since some
            # "os" module functions (which are called with this) do strict nominal type checking.
            late_binding=False,
        )

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        configuration.INSTALLED_APPS += ['django.contrib.staticfiles']


class WhitenoiseStaticFileMixin(StaticFileMixin):
    """
    Configure static file serving with Whitenoise for development and production.

    The package `whitenoise[brotli]` must be installed.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        # Insert immediately before staticfiles app
        staticfiles_index = configuration.INSTALLED_APPS.index('django.contrib.staticfiles')
        configuration.INSTALLED_APPS.insert(staticfiles_index, 'whitenoise.runserver_nostatic')

        # Insert immediately after SecurityMiddleware
        try:
            security_index = configuration.MIDDLEWARE.index(
                'django.middleware.security.SecurityMiddleware'
            )
        except ValueError:
            raise Exception(
                'WhitenoiseStaticFileMixin must be loaded after '
                'the SecurityMiddleware is added to MIDDLEWARE.'
            )
        configuration.MIDDLEWARE.insert(
            security_index + 1, 'whitenoise.middleware.WhiteNoiseMiddleware'
        )

    STATICFILES_STORAGE = 'whitenoise.storage.CompressedManifestStaticFilesStorage'



from configurations import values

from ._base import ConfigMixin


class _StorageMixin(ConfigMixin):
    """Abstract base for storage configs."""

    pass
    # For unity, subclasses should use "environ_name='STORAGE_BUCKET_NAME'" for
    # whatever particular setting is used to store the bucket name


class MinioStorageMixin(_StorageMixin):
    """
    Configure MinioMediaStorage.

    The following environment variables must be externally set:
    * DJANGO_MINIO_STORAGE_ACCESS_KEY
    * DJANGO_MINIO_STORAGE_SECRET_KEY
    * DJANGO_STORAGE_BUCKET_NAME

    This requires the `django-minio-storage` package to be installed.
    """

    DEFAULT_FILE_STORAGE = 'minio_storage.storage.MinioMediaStorage'
    MINIO_STORAGE_ENDPOINT = values.Value('localhost:9000')
    MINIO_STORAGE_USE_HTTPS = values.BooleanValue(False)
    MINIO_STORAGE_ACCESS_KEY = values.SecretValue()
    MINIO_STORAGE_SECRET_KEY = values.SecretValue()
    MINIO_STORAGE_MEDIA_BUCKET_NAME = values.Value(
        environ_name='STORAGE_BUCKET_NAME', environ_required=True
    )
    MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET = True
    MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY = 'READ_WRITE'
    MINIO_STORAGE_MEDIA_USE_PRESIGNED = True


class S3StorageMixin(_StorageMixin):
    """
    Configure S3Boto3Storage.

    The following environment variables must be externally set:
    * AWS_DEFAULT_REGION
    * AWS_ACCESS_KEY_ID
    * AWS_SECRET_ACCESS_KEY
    * DJANGO_STORAGE_BUCKET_NAME

    This requires the `django-storages[boto3]` package to be installed.
    """

    DEFAULT_FILE_STORAGE = 'storages.backends.s3boto3.S3Boto3Storage'

    # This exact environ_name is important, as direct use of Boto will also use it
    AWS_S3_REGION_NAME = values.Value(
        environ_prefix=None, environ_name='AWS_DEFAULT_REGION', environ_required=True
    )
    AWS_S3_ACCESS_KEY_ID = values.Value(
        environ_prefix=None, environ_name='AWS_ACCESS_KEY_ID', environ_required=True
    )
    AWS_S3_SECRET_ACCESS_KEY = values.Value(
        environ_prefix=None, environ_name='AWS_SECRET_ACCESS_KEY', environ_required=True
    )

    AWS_STORAGE_BUCKET_NAME = values.Value(
        environ_name='STORAGE_BUCKET_NAME', environ_required=True
    )

    # It's critical to use the v4 signature;
    # it isn't the upstream default only for backwards compatability reasons.
    AWS_S3_SIGNATURE_VERSION = 's3v4'

    AWS_S3_MAX_MEMORY_SIZE = 5 * 1024 * 1024
    AWS_S3_FILE_OVERWRITE = False
    AWS_QUERYSTRING_EXPIRE = 3600 * 6  # 6 hours



from typing import Type

from ._base import ComposedConfiguration, ConfigMixin


class FilterMixin(ConfigMixin):
    """
    Configure Django-filter.

    This requires the `django-filter` package to be installed.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        configuration.INSTALLED_APPS += ['django_filters']



from typing import Type

from configurations import values

from ._base import ComposedConfiguration, ConfigMixin


class SentryMixin(ConfigMixin):
    """
    Configure Sentry.io error reporting.

    This requires the `sentry-sdk` package to be installed.

    The `DJANGO_SENTRY_DSN` environment variable should be externally set to a Sentry DSN.

    The `DJANGO_SENTRY_ENVIRONMENT` and `DJANGO_SENTRY_RELEASE` environment variables may also
    be set, if desired.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        configuration.INSTALLED_APPS += [
            'composed_configuration.sentry.apps.SentryConfig',
        ]

    SENTRY_DSN = values.Value(None)

    SENTRY_ENVIRONMENT = values.Value(None)

    SENTRY_RELEASE = values.Value(None)



from typing import Type

from ._base import ComposedConfiguration, ConfigMixin


class AllauthMixin(ConfigMixin):
    """
    Configure Django Allauth.

    This requires the django-allauth package to be installed.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        configuration.INSTALLED_APPS += [
            'django.contrib.sites',
            'allauth',
            'allauth.account',
            'allauth.socialaccount',
        ]

        # girder_style should come before others, to ensure its template overrides are found
        configuration.INSTALLED_APPS.insert(0, 'girder_style')

    # The sites framework requires this to be set.
    # In the unlikely case where a database's pk sequence for the django_site table is not reset,
    # the default site object could have a different pk. Then this will need to be overridden
    # downstream.
    SITE_ID = 1

    AUTHENTICATION_BACKENDS = [
        # Django's built-in ModelBackend is not necessary, since all users will be
        # authenticated by their email address
        'allauth.account.auth_backends.AuthenticationBackend',
    ]

    # see configuration documentation at
    #   https://django-allauth.readthedocs.io/en/latest/configuration.html

    # Require email verification, but this can be overridden
    ACCOUNT_EMAIL_VERIFICATION = 'mandatory'

    # Make Django and Allauth redirects consistent, but both may be overridden
    LOGIN_REDIRECT_URL = '/'
    ACCOUNT_LOGOUT_REDIRECT_URL = '/'

    # Use email as the identifier for login
    ACCOUNT_AUTHENTICATION_METHOD = 'email'
    ACCOUNT_EMAIL_REQUIRED = True
    ACCOUNT_USERNAME_REQUIRED = False

    # Set the username as the email
    ACCOUNT_ADAPTER = 'composed_configuration.authentication.adapter.EmailAsUsernameAccountAdapter'
    ACCOUNT_USER_MODEL_USERNAME_FIELD = None

    # Quality of life improvements, but may not work if the browser is closed
    ACCOUNT_SESSION_REMEMBER = True
    ACCOUNT_LOGIN_ON_EMAIL_CONFIRMATION = True
    ACCOUNT_LOGIN_ON_PASSWORD_RESET = True

    # These will permit GET requests to mutate the user state, but significantly improve usability
    ACCOUNT_LOGOUT_ON_GET = True
    ACCOUNT_CONFIRM_EMAIL_ON_GET = True

    # This will likely become the default in the future, but enable it now
    ACCOUNT_PRESERVE_USERNAME_CASING = False



from typing import Type

from configurations import values

from ._base import ComposedConfiguration, ConfigMixin


class DatabaseMixin(ConfigMixin):
    """
    Configure a single PostgreSQL database.

    The `DJANGO_DATABASE_URL` environment variable must be externally set
    to a PostgreSQL URL including credentials and the database name.

    This requires the `psycopg2` package to be installed.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        configuration.INSTALLED_APPS += ['django.contrib.postgres']

    # This cannot have a default value, since the password and database
    # name are always set by the service admin.
    DATABASES = values.DatabaseURLValue(
        environ_name='DATABASE_URL',
        # django-configurations has environ_prefix=None by default here
        environ_prefix='DJANGO',
        environ_required=True,
        # Additional kwargs to DatabaseURLValue are passed to dj-database-url,
        # then passed through to the Django database options.
        engine='django.db.backends.postgresql',
        conn_max_age=600,
    )


class HerokuDatabaseMixin(DatabaseMixin):
    """
    Configure a single Heroku PostgreSQL database.

    The `DATABASE_URL` environment variable must be externally set
    to a PostgreSQL URL including credentials and the database name.
    """

    # Heroku sets the environment variable as DATABASE_URL, so drop the
    # DJANGO_ prefix.
    DATABASES = values.DatabaseURLValue(
        environ_name='DATABASE_URL',
        environ_prefix=None,
        environ_required=True,
        # Additional kwargs here.
        engine='django.db.backends.postgresql',
        conn_max_age=600,
        # Heroku is expected to always provide SSL.
        ssl_require=True,
    )



from typing import Type

from ._base import ComposedConfiguration, ConfigMixin


class DebugMixin(ConfigMixin):
    """
    Configure debug tooling.

    This requires the `django-debug-toolbar` package to be installed.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        configuration.INSTALLED_APPS += ['debug_toolbar']

        # Include Debug Toolbar middleware as early as possible in the list.
        # However, it must come after any other middleware that encodes the response’s content,
        # such as GZipMiddleware.
        configuration.MIDDLEWARE.insert(0, 'debug_toolbar.middleware.DebugToolbarMiddleware')



from typing import Dict, Type, cast

from configurations import values

from ._base import ComposedConfiguration, ConfigMixin


class _EmailMixin(ConfigMixin):
    """Abstract base for email configs."""

    pass


class SmtpEmailMixin(_EmailMixin):
    """
    Configure Django's email sending.

    The following environment variables must be externally set:
    * `DJANGO_EMAIL_URL`, as a URL for login to an STMP server, as parsed by `dj-email-url`. This
      typically will start with `submission:`. Special characters in passwords must be URL-encoded.
      See https://pypi.org/project/dj-email-url/ for full details.
    * `DJANGO_DEFAULT_FROM_EMAIL`, as the default From address for outgoing email.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        email = cast(
            Dict[str, str],
            values.EmailURLValue(
                environ_name='EMAIL_URL',
                environ_prefix='DJANGO',
                environ_required=True,
                # Disable late_binding, to make this return a usable value (which is a simple dict)
                # immediately
                late_binding=False,
            ),
        )
        for email_setting, email_setting_value in email.items():
            setattr(configuration, email_setting, email_setting_value)

    # Set both settings from DJANGO_DEFAULT_FROM_EMAIL
    DEFAULT_FROM_EMAIL = values.EmailValue(environ_required=True)
    SERVER_EMAIL = values.EmailValue(environ_name='DEFAULT_FROM_EMAIL', environ_required=True)


class ConsoleEmailMixin(_EmailMixin):
    EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'



from typing import Type

from configurations import Configuration, values
from configurations.base import ConfigurationBase

# With the default "late_binding=False", and "environ_name" is specified or "environ=False",
# even Values from non-included classes (e.g. `AWS_DEFAULT_REGION) get immediately evaluated and
# expect env vars to be set. Also, immediately evaluated Values cannot be tweaked effectively
# in "before_binding").
values.Value.late_binding = True


# Fix https://github.com/jazzband/django-configurations/issues/263
# Since this attribute is set by the metaclass, it exists on the already-instantied Configuration
# class. It also will be re-set on every subclass, but since those are not instantied yet, we
# can replace the metaclass for all subclasses with a fixed one.
try:
    del Configuration.DEFAULT_HASHING_ALGORITHM
except AttributeError:
    pass


class FixedConfigurationBase(ConfigurationBase):
    def __new__(cls, name, bases, attrs):
        obj = super().__new__(cls, name, bases, attrs)
        try:
            del obj.DEFAULT_HASHING_ALGORITHM
        except AttributeError:
            pass
        return obj


class ComposedConfiguration(Configuration, metaclass=FixedConfigurationBase):
    """
    Abstract base for composed Configuration.

    This must always be specified as a base class after Config mixins.
    """

    @classmethod
    def pre_setup(cls):
        super().pre_setup()

        # For every class in the inheritance hierarchy
        # Reverse order allows more base classes to run first
        for base_cls in reversed(cls.__mro__):
            # If the class has "before_binding" as its own (non-inherited) method
            if 'before_binding' in base_cls.__dict__:
                base_cls.before_binding(cls)


class ConfigMixin:
    """Abstract mixin for composable Config sections."""

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        """
        Run before values are fully bound with environment variables.

        `configuration` refers to the final Configuration class, so settings from
        other Configs in the final hierarchy may be referenced.
        """
        pass



from ._base import ConfigMixin


def _filter_favicon_requests(record):
    if (
        record.name == 'django.request'
        and hasattr(record, 'request')
        and record.request.path == '/favicon.ico'
    ):
        return False

    if (
        record.name == 'django.server'
        and len(record.args) >= 1
        and str(record.args[0]).startswith('GET /favicon.ico ')
    ):
        return False

    return True


def _filter_static_requests(record):
    if (
        record.name == 'django.server'
        and len(record.args) >= 1
        and str(record.args[0]).startswith('GET /static/')
    ):
        return False

    return True


class LoggingMixin(ConfigMixin):
    """
    Configure Django logging.

    This requires the `rich` package to be installed.
    """

    LOGGING = {
        'version': 1,
        # Replace existing logging configuration
        'incremental': False,
        # This redefines all of Django's declared loggers, but most loggers are implicitly
        # declared on usage, and should not be disabled. They often propagate their output
        # to the root anyway.
        'disable_existing_loggers': False,
        'formatters': {'rich': {'datefmt': '[%X]'}},
        'filters': {
            'filter_favicon_requests': {
                '()': 'django.utils.log.CallbackFilter',
                'callback': _filter_favicon_requests,
            },
            'filter_static_requests': {
                '()': 'django.utils.log.CallbackFilter',
                'callback': _filter_static_requests,
            },
        },
        'handlers': {
            'console': {
                'class': 'rich.logging.RichHandler',
                'formatter': 'rich',
                'filters': ['filter_favicon_requests', 'filter_static_requests'],
            },
        },
        # Existing loggers actually contain direct (non-string) references to existing handlers,
        # so after redefining handlers, all existing loggers must be redefined too
        'loggers': {
            # Configure the root logger to output to the console
            '': {'level': 'INFO', 'handlers': ['console'], 'propagate': False},
            # Django defines special configurations for the "django" and "django.server" loggers,
            # but we will manage all content at the root logger instead, so reset those
            # configurations.
            'django': {
                'handlers': [],
                'level': 'NOTSET',
                'propagate': True,
            },
            'django.server': {
                'handlers': [],
                'level': 'NOTSET',
                'propagate': True,
            },
        },
    }



from typing import NoReturn, Type

from configurations import values

from ._base import ComposedConfiguration, ConfigMixin


class DjangoMixin(ConfigMixin):
    """
    Configure a basic Django project.

    Downstreams must explicitly define the settings:
    * `WSGI_APPLICATION`, as a string with the Python import path of the WSGI `application` in the
      `wsgi.py` file.
    * `ROOT_URLCONF`, as a string with the Python import path of the base `urls.py` file.

    The following environment variables must be externally set:
    * `SECRET_KEY`, as a random string.
    * `DJANGO_ALLOWED_HOSTS`, as a comma-delimited list of fully-qualified domain names from
       which this server will be accessed.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        # These are often extended, so update their values to avoid fragility due to mixin ordering
        configuration.INSTALLED_APPS += [
            'django.contrib.admin',
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.messages',
        ]
        configuration.MIDDLEWARE += [
            'django.middleware.security.SecurityMiddleware',
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.common.CommonMiddleware',
            'django.middleware.csrf.CsrfViewMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
            'django.middleware.clickjacking.XFrameOptionsMiddleware',
        ]

    SECRET_KEY = values.SecretValue()
    ALLOWED_HOSTS = values.ListValue(environ_required=True)

    @property
    def WSGI_APPLICATION(self) -> NoReturn:  # noqa: N802
        raise Exception('WSGI_APPLICATION must be explicitly set.')

    @property
    def ROOT_URLCONF(self) -> NoReturn:  # noqa: N802
        raise Exception('ROOT_URLCONF must be explicitly set.')

    TEMPLATES = [
        {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': [],
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                    'django.contrib.messages.context_processors.messages',
                ],
            },
        },
    ]

    # Until https://github.com/pennersr/django-allauth/issues/2826 is fixed and released,
    # this must be AutoField
    DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

    # Password validation
    # https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators
    AUTH_PASSWORD_VALIDATORS = [
        {'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator'},
        {'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator'},
        {'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator'},
        {'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator'},
    ]

    # Internationalization
    # https://docs.djangoproject.com/en/3.0/topics/i18n/
    LANGUAGE_CODE = 'en-us'
    USE_TZ = True
    TIME_ZONE = 'UTC'
    USE_I18N = True
    USE_L10N = True



from typing import Type

from configurations import values

from ._base import ComposedConfiguration, ConfigMixin


class CorsMixin(ConfigMixin):
    """
    Configure CORS headers.

    The following environment variables may be externally set:
    * `DJANGO_CORS_ORIGIN_WHITELIST`
    * `DJANGO_CORS_ORIGIN_REGEX_WHITELIST`

    This requires the `django-cors-headers` package to be installed.
    This also must be loaded after `WhitenoiseStaticFileMixin`.
    """

    @staticmethod
    def before_binding(configuration: Type[ComposedConfiguration]) -> None:
        configuration.INSTALLED_APPS += ['corsheaders']

        # CorsMiddleware must be added immediately before WhiteNoiseMiddleware, so this can
        # potentially add CORS headers to those responses too.
        # Accordingly, CorsMixin must be loaded after WhitenoiseStaticFileMixin, so it can
        # find the existing entry and insert accordingly.
        try:
            whitenoise_index = configuration.MIDDLEWARE.index(
                'whitenoise.middleware.WhiteNoiseMiddleware'
            )
        except ValueError:
            raise Exception('CorsConfig must be loaded after WhitenoiseStaticFileMixin.')
        configuration.MIDDLEWARE.insert(whitenoise_index, 'corsheaders.middleware.CorsMiddleware')

    CORS_ORIGIN_WHITELIST = values.ListValue()
    CORS_ORIGIN_REGEX_WHITELIST = values.ListValue()



import logging

from django.apps import AppConfig
from django.conf import settings
import sentry_sdk
from sentry_sdk.integrations.celery import CeleryIntegration
from sentry_sdk.integrations.django import DjangoIntegration
from sentry_sdk.integrations.logging import LoggingIntegration


class SentryConfig(AppConfig):
    name = 'composed_configuration.sentry'
    verbose_name = 'Composed configuration Sentry support'

    def ready(self):
        sentry_sdk.init(
            # If a "dsn" is not explicitly passed, sentry_sdk will attempt to find the DSN in
            # the SENTRY_DSN environment variable; however, by pulling it from an explicit setting,
            # it can be overridden by downstream project settings.
            dsn=settings.SENTRY_DSN,
            environment=settings.SENTRY_ENVIRONMENT,
            release=settings.SENTRY_RELEASE,
            integrations=[
                LoggingIntegration(level=logging.INFO, event_level=logging.WARNING),
                DjangoIntegration(),
                CeleryIntegration(),
            ],
            # Send traces for non-exception events too
            attach_stacktrace=True,
            # Submit request User info from Django
            send_default_pii=True,
        )



from pathlib import Path


def _is_docker() -> bool:
    """Determine whether the current environment is within a Docker container."""
    # https://tuhrig.de/how-to-know-you-are-inside-a-docker-container/
    cgroup_file = Path('/proc/self/cgroup')
    if cgroup_file.exists():
        with cgroup_file.open() as cgroup_stream:
            # This file should be small enough to fully read into memory
            return 'docker' in cgroup_stream.read()
    else:
        # Maybe running on macOS
        return False


class _AlwaysContains(object):
    """An object which always returns True for `x in _AlwaysContains()` operations."""

    def __contains__(self, item) -> bool:
        # https://stackoverflow.com/a/49818040
        return True



from pathlib import Path

from setuptools import find_packages, setup

readme_file = Path(__file__).parent / 'README.md'
with readme_file.open() as f:
    long_description = f.read()

_base_extras = [
    'django>=3',
    'django-allauth',
    'django-cors-headers',
    'django-extensions',
    'django-filter',
    'django-girder-style',
    'django-girder-utils',
    # django-oauth-toolkit==1.3.3 doesn't have working PKCE
    'django-oauth-toolkit>=1.4.0',
    'drf-yasg',
    'psycopg2',
    'rich',
    'whitenoise[brotli]',
]

setup(
    name='django-composed-configuration',
    description='Turnkey Django settings for data management applications.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    url='https://github.com/girder/django-composed-configuration',
    project_urls={
        'Bug Reports': 'https://github.com/girder/django-composed-configuration/issues',
        'Source': 'https://github.com/girder/django-composed-configuration',
    },
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    keywords='django girder configuration configurations setting settings',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django :: 3.0',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python',
    ],
    python_requires='>=3.8',
    install_requires=[
        'django-configurations[database,email]',
    ],
    extras_require={
        # Required for DevelopmentBaseConfiguration / TestingBaseConfiguration
        'dev': _base_extras + ['django-debug-toolbar', 'django-minio-storage'],
        # Required for ProductionBaseConfiguration
        'prod': _base_extras
        + [
            'django-storages[boto3]',
            'sentry-sdk',
        ],
    },
    packages=find_packages(),
    include_package_data=True,
)



from allauth.account.adapter import DefaultAccountAdapter


class EmailAsUsernameAccountAdapter(DefaultAccountAdapter):
    """Automatically populate the username as the email address."""

    def populate_username(self, request, user):
        user.username = user.email
